@@ -1,6 +1,6 @@
 /*
   Mode switching tool for controlling mode of 'multi-state' USB devices
-  Version 2.5.0, 2017/01/17
+  Version 2.5.1, 2017/08/06
 
   Copyright (C) 2007 - 2017 Josua Dietze (mail to "usb_admin" at the domain
   of the home page; or write a personal message through the forum to "Josh".
@@ -45,7 +45,7 @@
 
 /* Recommended tab size: 4 */
 
-#define VERSION "2.5.0"
+#define VERSION "2.5.1"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -139,6 +139,7 @@ unsigned int ModeMap = 0;
 #define PANTECH_MODE		0x00001000
 #define HUAWEINEW_MODE		0x00002000
 #define OPTION_MODE			0x00004000
+#define HUAWEIALT_MODE		0x00008000
 
 
 int PantechMode=0;
@@ -158,7 +159,7 @@ char **Messages = NULL;
 FILE *output;
 
 
-/* Settable Interface and Configuration (for debugging mostly) (jmw) */
+/* Settable interface, altsetting (for debugging mostly) and configuration */
 int Interface = -1, Configuration = 0, AltSetting = -1;
 
 
@@ -181,6 +182,7 @@ static struct option long_options[] = {
 	{"detach-only",			no_argument, 0, 'd'},
 	{"huawei-mode",			no_argument, 0, 'H'},
 	{"huawei-new-mode",		no_argument, 0, 'J'},
+	{"huawei-alt-mode",		no_argument, 0, 'X'},
 	{"sierra-mode",			no_argument, 0, 'S'},
 	{"sony-mode",			no_argument, 0, 'O'},
 	{"qisda-mode",			no_argument, 0, 'B'},
@@ -223,6 +225,7 @@ void readConfigFile(const char *configFilename)
 	ParseParamBoolMap(configFilename, DetachStorageOnly, ModeMap, DETACHONLY_MODE);
 	ParseParamBoolMap(configFilename, HuaweiMode, ModeMap, HUAWEI_MODE);
 	ParseParamBoolMap(configFilename, HuaweiNewMode, ModeMap, HUAWEINEW_MODE);
+	ParseParamBoolMap(configFilename, HuaweiAltMode, ModeMap, HUAWEIALT_MODE);
 	ParseParamBoolMap(configFilename, SierraMode, ModeMap, SIERRA_MODE);
 	ParseParamBoolMap(configFilename, SonyMode, ModeMap, SONY_MODE);
 	ParseParamBoolMap(configFilename, GCTMode, ModeMap, GCT_MODE);
@@ -283,6 +286,8 @@ void printConfig()
 		fprintf (output,"HuaweiMode=1\n");
 	if (ModeMap & HUAWEINEW_MODE)
 		fprintf (output,"HuaweiNewMode=1\n");
+	if (ModeMap & HUAWEIALT_MODE)
+		fprintf (output,"HuaweiAltMode=1\n");
 	if (ModeMap & SIERRA_MODE)
 		fprintf (output,"SierraMode=1\n");
 	if (ModeMap & SONY_MODE)
@@ -343,7 +348,7 @@ int readArguments(int argc, char **argv)
 
 	while (1)
 	{
-		c = getopt_long (argc, argv, "hejWQDndKHJSOBEGTNALZUF:RItv:p:V:P:C:m:M:2:3:w:r:c:i:u:a:s:f:b:g:",
+		c = getopt_long (argc, argv, "hejWQDndKHJSOBEGTNALZUXF:RItv:p:V:P:C:m:M:2:3:w:r:c:i:u:a:s:f:b:g:",
 					long_options, &option_index);
 
 		/* Detect the end of the options. */
@@ -369,6 +374,7 @@ int readArguments(int argc, char **argv)
 			case 'd': ModeMap = ModeMap + DETACHONLY_MODE; break;
 			case 'H': ModeMap = ModeMap + HUAWEI_MODE; break;
 			case 'J': ModeMap = ModeMap + HUAWEINEW_MODE; break;
+			case 'X': ModeMap = ModeMap + HUAWEIALT_MODE; break;
 			case 'S': ModeMap = ModeMap + SIERRA_MODE; break;
 			case 'O': ModeMap = ModeMap + SONY_MODE; break;; break;
 			case 'B': ModeMap = ModeMap + QISDA_MODE; break;
@@ -429,7 +435,7 @@ int readArguments(int argc, char **argv)
 int main(int argc, char **argv)
 {
 	int ret=0, numDefaults=0, sonySuccess=0, i;
-	int currentConfigVal=0, defaultClass=0, interfaceClass=0;
+	int defaultClass=0, interfaceClass=0, currentConfigVal=0;
 	struct libusb_device_descriptor descriptor;
 	enum libusb_error libusbError;
 
@@ -598,8 +604,8 @@ int main(int argc, char **argv)
 		}
 
 	if ((strlen(MessageContent) && strncmp("55534243",MessageContent,8) == 0)
-			|| StandardEject || ModeMap & HUAWEINEW_MODE || ModeMap & CISCO_MODE
-			|| ModeMap & OPTION_MODE)
+			|| StandardEject || ModeMap & HUAWEINEW_MODE || ModeMap & HUAWEIALT_MODE
+			|| ModeMap & CISCO_MODE || ModeMap & OPTION_MODE)
 		if (defaultClass != 8) {
 			fprintf(stderr, "Error: can't use storage command in MessageContent with interface %d; "
 				"interface class is %d, expected 8. Abort\n\n", Interface, defaultClass);
@@ -608,7 +614,8 @@ int main(int argc, char **argv)
 
 	/* Check or get endpoints and alloc message list if needed*/
 	if (strlen(MessageContent) || StandardEject || ModeMap & CISCO_MODE
-				|| ModeMap & HUAWEINEW_MODE || ModeMap & OPTION_MODE) {
+				|| ModeMap & HUAWEINEW_MODE || ModeMap & HUAWEIALT_MODE
+				|| ModeMap & OPTION_MODE) {
 
 		Messages = (char**) calloc(MSG_DIM, sizeof(char*));
 		for (i = 0; i < MSG_DIM; i++) {
@@ -633,7 +640,7 @@ int main(int argc, char **argv)
 
 	}
 
-	if (show_progress) {
+	if (verbose) {
 		fprintf(output,"\nUSB description data (for identification)\n");
 		deviceDescription();
 	}
@@ -672,7 +679,7 @@ int main(int argc, char **argv)
 
 	if (ModeMap & DETACHONLY_MODE) {
 		SHOW_PROGRESS(output,"Detach storage driver as switching method ...\n");
-		ret = detachDriver();
+		ret = detachDrivers();
 		if (ret == 2)
 			SHOW_PROGRESS(output," You may want to remove the storage driver manually\n");
 	}
@@ -684,14 +691,14 @@ int main(int argc, char **argv)
 		switchSierraMode();
 	}
 	if(ModeMap & GCT_MODE) {
-		detachDriver();
+		detachDrivers();
 		switchGCTMode();
 	}
 	if(ModeMap & QISDA_MODE) {
 		switchQisdaMode();
 	}
 	if(ModeMap & KOBIL_MODE) {
-		detachDriver();
+		detachDrivers();
 		switchKobilMode();
 	}
 	if(ModeMap & QUANTA_MODE) {
@@ -704,15 +711,15 @@ int main(int argc, char **argv)
 		switchActionMode();
 	}
 	if(ModeMap & CISCO_MODE) {
-		detachDriver();
+		detachDrivers();
 		switchCiscoMode();
 	}
 	if(ModeMap & BLACKBERRY_MODE) {
-		detachDriver();
+		detachDrivers();
 	    switchBlackberryMode();
 	}
 	if(ModeMap & PANTECH_MODE) {
-		detachDriver();
+		detachDrivers();
 		if (PantechMode > 1)
 			switchPantechMode();
 		else
@@ -727,7 +734,7 @@ int main(int argc, char **argv)
 
 	if (StandardEject) {
 		SHOW_PROGRESS(output,"Sending standard EJECT sequence\n");
-		detachDriver();
+		detachDrivers();
 
 		strcpy(Messages[0],"5553424387654321000000000000061e000000000000000000000000000000");
 		strcpy(Messages[1],"5553424397654321000000000000061b000000020000000000000000000000");
@@ -739,17 +746,21 @@ int main(int argc, char **argv)
 		switchSendMessage();
 	} else if (ModeMap & HUAWEINEW_MODE) {
 		SHOW_PROGRESS(output,"Using standard Huawei switching message\n");
-		detachDriver();
+		detachDrivers();
 		strcpy(Messages[0],"55534243123456780000000000000011062000000101000100000000000000");
 		switchSendMessage();
+	} else if (ModeMap & HUAWEIALT_MODE) {
+		SHOW_PROGRESS(output,"Using alternative Huawei switching message\n");
+		detachDrivers();
+		strcpy(Messages[0],"55534243123456780000000000000011063000000000010000000000000000");
+		switchSendMessage();
 	} else if (ModeMap & OPTION_MODE) {
 		SHOW_PROGRESS(output,"Using standard Option switching message\n");
-		detachDriver();
-//		strcpy(MessageContent,"55534243123456780100000080000601000000000000000000000000000000");
+		detachDrivers();
 		strcpy(Messages[0],"55534243123456780000000000000601000000000000000000000000000000");
 		switchSendMessage();
 	} else if (strlen(MessageContent)) {
-		detachDriver();
+		detachDrivers();
 		strcpy(Messages[0],MessageContent);
 		switchSendMessage();
 	}
@@ -765,7 +776,9 @@ int main(int argc, char **argv)
 				}
 			}
 		} else {
-			SHOW_PROGRESS(output,"Target configuration %d found. Do nothing\n", currentConfigVal);
+			SHOW_PROGRESS(output,"Target configuration %d already active. Nothing to do. Bye!\n\n", currentConfigVal);
+			close_all();
+			exit(0);
 		}
 	}
 
@@ -1006,13 +1019,14 @@ retry:
 			// UFI command
 			SHOW_PROGRESS(output,"Read the response to message %d (CSW) ...\n", i+1);
 			ret = read_bulk(ResponseEndpoint, ByteString, 13);
-			if (ret >= 0)
+			if (ret >= 0) {
 				SHOW_PROGRESS(output,", status %d",ByteString[12]);
-		} else {
+			}
+		} /* else {
 			// Other bulk transfer
 			SHOW_PROGRESS(output,"Read the response to message %d ...\n", i+1);
 			ret = read_bulk(ResponseEndpoint, ByteString, strlen(Messages[i])/2 );
-		}
+		}*/
 		SHOW_PROGRESS(output,"\n");
 		if (ret == LIBUSB_TRANSFER_STALL && retries--) {
 			SHOW_PROGRESS(output,"Endpoint stalled. Resetting ...\n");
@@ -1052,14 +1066,18 @@ skip:
 
 int switchConfiguration ()
 {
-	int ret, count = SWITCH_CONFIG_MAXTRIES;
+	int ret;
 
 	SHOW_PROGRESS(output,"Change configuration to %i ...\n", Configuration);
-	while (((ret = libusb_set_configuration(devh, Configuration)) < 0) && --count) {
-		SHOW_PROGRESS(output," Device is busy, try to detach kernel driver\n");
-		detachDriver();
+	detachDrivers();
+	ret = libusb_set_configuration(devh, -1);
+	if (ret < 0) {
+		SHOW_PROGRESS(output," Resetting the configuration failed (error %d). Try to continue\n", ret);
 	}
-	if (ret < 0 ) {
+	/* Empirically tested wait period, improves reliability of configuration change */
+	usleep(100000);
+	ret = libusb_set_configuration(devh, Configuration);
+	if (ret < 0) {
 		SHOW_PROGRESS(output," Changing the configuration failed (error %d). Try to continue\n", ret);
 		return 0;
 	} else {
@@ -1374,7 +1392,7 @@ skip:
 int switchSonyMode ()
 {
 	int ret, i, found;
-	detachDriver();
+	detachDrivers();
 
 	if (CheckSuccess) {
 		CheckSuccess = 0;
@@ -1441,34 +1459,41 @@ int switchSonyMode ()
 
 /* Detach driver
  */
-int detachDriver()
+int detachDrivers()
 {
 
-	int ret;
-	SHOW_PROGRESS(output,"Looking for active driver ...\n");
+	int i, ret;
+	SHOW_PROGRESS(output,"Looking for active drivers ...\n");
 	ret = libusb_kernel_driver_active(devh, 0);
 	if (ret == LIBUSB_ERROR_NOT_SUPPORTED) {
 		fprintf(output," Can't do driver detection on this platform.\n");
 		return 2;
 	}
-	if (ret < 0) {
-		fprintf(output," Driver check failed with error %d. Try to continue\n", ret);
-		return 2;
-	}
-	if (ret == 0) {
-		SHOW_PROGRESS(output," No active driver found. Detached before or never attached\n");
-		return 1;
-	}
 
-	ret = libusb_detach_kernel_driver(devh, Interface);
-	if (ret == LIBUSB_ERROR_NOT_SUPPORTED) {
-		fprintf(output," Can't do driver detaching on this platform.\n");
-		return 2;
+	struct libusb_config_descriptor *config;
+	libusb_get_active_config_descriptor(dev, &config);
+
+	for (i=0; i<config->bNumInterfaces; i++) {
+		ret = libusb_kernel_driver_active(devh, i);
+		if (ret < 0) {
+			SHOW_PROGRESS(output," Failed to check driver status for interface %d (error %d)\n Try to continue\n",i,ret);
+			continue;
+		}
+		if (ret) {
+			ret = libusb_detach_kernel_driver(devh, i);
+			if (ret == LIBUSB_ERROR_NOT_SUPPORTED) {
+				fprintf(output," Can't do driver detaching on this platform.\n");
+				return 2;
+			}
+			if (ret == 0) {
+				SHOW_PROGRESS(output," OK, driver detached\n");
+			} else {
+				SHOW_PROGRESS(output," Driver detach failed for interface %d (error %d).\n Try to continue\n",i,ret);
+				continue;
+			}
+		}
 	}
-	if (ret == 0) {
-		SHOW_PROGRESS(output," OK, driver detached\n");
-	} else
-		SHOW_PROGRESS(output," Driver detach failed (error %d). Try to continue\n", ret);
+	libusb_free_config_descriptor(config);
 	return 1;
 }
 
@@ -1732,7 +1757,8 @@ struct libusb_device* search_devices( int *numFound, int vendor, char* productLi
 			product <<= 8;
 			product += (unsigned char)buffer[1];
 			if (product == descriptor.idProduct) {
-				SHOW_PROGRESS(output,"   product ID matched\n");
+				if (verbose)
+					fprintf(output,"   product ID matched\n");
 
 				if (targetClass != 0) {
 					// TargetClass is set, check class of first interface
@@ -1770,7 +1796,7 @@ struct libusb_device* search_devices( int *numFound, int vendor, char* productLi
 								fprintf (output,"   count device\n");
 						}
 					}
-				} else if (configuration > 0) {
+/*				} else if (configuration > 0) {
 					// Configuration parameter is set, check device configuration
 					int testconfig = get_current_config_value(dev);
 					if (testconfig != configuration) {
@@ -1785,6 +1811,7 @@ struct libusb_device* search_devices( int *numFound, int vendor, char* productLi
 					} else
 						if (verbose)
 							fprintf (output,"   device not counted, target configuration reached\n");
+*/
 				} else {
 					// Neither TargetClass nor Configuration are set
 					(*numFound)++;
@@ -2082,6 +2109,7 @@ void printHelp()
 	" -d, --detach-only             detach the active driver, no further action\n"
 	" -H, --huawei-mode             apply a special procedure\n"
 	" -J, --huawei-new-mode         apply a special procedure\n"
+	" -X, --huawei-alt-mode         apply a special procedure\n"
 	" -S, --sierra-mode             apply a special procedure\n"
 	" -O, --sony-mode               apply a special procedure\n"
 	" -G, --gct-mode                apply a special procedure\n"
